__version__ = '0.2.9'

#These are brought into the behaveml namespace
from behaveml.io import read_DLC_tracks, save_DLC_tracks_h5, load_sklearn_model, save_sklearn_model
from behaveml.interpolation import interpolate_lowconf_points
from behaveml.video import VideosetDataFrame, clone_metadata, load_videodataset

from behaveml.features import cnn_probability_feature_maker, mars_feature_maker, social_feature_maker, \
                              distance_feature_maker, speed_feature_maker, marsreduced_feature_maker, \
                              com_interanimal_feature_maker, com_interanimal_speed_feature_maker, \
                              com_feature_maker, com_velocity_feature_maker

from behaveml.unsupervised import compute_tsne_embedding

#from .version import __version__

#These are the functions imported when doing 'from behaveml import *'
__all__ = [VideosetDataFrame, clone_metadata, read_DLC_tracks, load_videodataset, 
           load_sklearn_model, save_sklearn_model, save_DLC_tracks_h5, 
           mars_feature_maker, cnn_probability_feature_maker, 
           distance_feature_maker, speed_feature_maker, social_feature_maker,
           marsreduced_feature_maker, interpolate_lowconf_points, compute_tsne_embedding,
           com_interanimal_feature_maker, com_interanimal_speed_feature_maker, \
           com_feature_maker, com_velocity_feature_maker, \
           distance_feature_maker]


from sklearn.preprocessing import StandardScaler
from sklearn.manifold import TSNE
import numpy as np
from scipy import signal
from tqdm import tqdm
from sklearn.neighbors import KernelDensity
import umap
from skimage.segmentation import watershed
from behaveml import VideosetDataFrame
import pandas as pd

def compute_tsne_embedding(dataset : VideosetDataFrame, 
                           cols : list, 
                           N_rows : int = 20000, 
                           n_components = 2, 
                           perplexity = 30) -> tuple:
    """Compute TSNE embedding. Only for a random subset of rows.
    
    Args:
        dataset: Input data
        cols: A list of column names to produce the embedding for
        N_rows: A number of rows to randomly sample for the embedding. Only these rows are embedded.
        n_components: The number of dimensions to embed the data into.
        perplexity: The perplexity of the TSNE embedding.

    Returns:
        The tuple: 
            - A numpy array with the embedding data, only for a random subset of row
            - The rows that were used for the embedding
    """
    
    tsne_data = StandardScaler().fit_transform(dataset.data[cols])
    random_indices = np.random.choice(tsne_data.shape[0], min(N_rows, tsne_data.shape[0]), replace = False)
    tsne_data = tsne_data[random_indices, :]
    tsne_embedding = TSNE(n_components=n_components, init = 'pca', perplexity = perplexity).fit_transform(tsne_data)
    return tsne_embedding, random_indices

def compute_morlet(data : np.ndarray, 
                   dt : float = 1/30, 
                   n_freq : int = 5, 
                   w : float = 3) -> np.ndarray:
    """ Compute morlet wavelet transform of a time series.
    
    Args:
        data: A 2D array containing the time series data, with dimensions (n_pts x n_channels)
        dt: The time step of the time series
        n_freq: The number of frequencies to compute
        w: The width of the morlet wavelet
        
    Returns
        A 2D numpy array with the morlet wavelet transform. The first dimension is the frequency, the second is the time.
    """
    if type(data) is pd.DataFrame:
        data = data.to_numpy()
    fs = 1/dt
    freq = np.geomspace(1, fs/2, n_freq)
    widths = w*fs / (2*freq*np.pi)
    morlet = []
    for idx in tqdm(range(data.shape[1])):
        morlet.append(np.abs(signal.cwt(data[:,idx], signal.morlet2, widths, w=w)))
    morlet = np.stack(morlet)
    return morlet    

def compute_density(dataset : VideosetDataFrame, 
                    embedding_extent : tuple, 
                    bandwidth : float = 0.5, 
                    n_pts : int = 300,
                    N_sample_rows : int = 50000,
                    rows : list = None) -> np.ndarray:
    """Compute kernel density estimate of embedding.
    
    Args:
        dataset: VideosetDataFrame with embedding data loaded in it. (Must have already populated columns named 'embedding_0', 'embedding_1')
        embedding_extent: the bounds in which to apply the density estimate. Has the form (xmin, xmax, ymin, ymax)
        bandwidth: the Gaussian kernel bandwidth. Will depend on the scale of the embedding. Can be changed to affect the number of clusters pulled out
        n_pts: number of points over which to evaluate the KDE
        N_sample_rows: number of rows to randomly sample to generate estimate
        rows: If provided, use these rows instead of a random sample

    Returns:
        Numpy array with KDE over the specified square region in the embedding space, with dimensions (n_pts x n_pts)    
    """

    xmin, xmax, ymin, ymax = embedding_extent
    den_est = KernelDensity(bandwidth = bandwidth)
    if rows is not None:
        sample_rows = rows
    else:
        sample_rows = np.random.choice(dataset.data.index, min(N_sample_rows, len(dataset.data)), replace = False)
    print("Fitting KDE")
    den_est.fit(dataset.data.loc[sample_rows, ['embedding_0', 'embedding_1']])
    X_plot = np.array(np.meshgrid(np.linspace(xmin, xmax, n_pts), np.linspace(ymin, ymax, n_pts))).T.reshape(-1, 2)
    print("Computing grid of densities")
    dens = den_est.score_samples(X_plot)
    dens = np.exp(dens)
    dens_matrix = dens.reshape(n_pts, n_pts).T
    return dens_matrix

def compute_watershed(dens_matrix : np.ndarray, 
                      positive_only : bool = False, 
                      cutoff : float = 0) -> tuple:
    """Compute watershed clustering of a density matrix. 
    
    Args:
        dens_matrix: A square 2D numpy array, output from compute_density, containing the kernel density estimate of the embedding.
        positive_only: Whether to apply a threshold, 'cutoff'. If applied, 'cutoff' is subtracted from dens_matrix, and any value below zero is set to zero. Useful for only focusing on high density clusters.
        cutoff: The cutoff value to apply if positive_only = True
        
    Returns:
        A numpy array with the same dimensions as dens_matrix. Each value in the array is the cluster ID for that coordinate.
    """

    if positive_only:
        dm = np.maximum(cutoff, dens_matrix)-cutoff
    else:
        dm = dens_matrix
    image = (255*dm/np.max(dm)).astype(int)
    labels = watershed(-dm, None, mask = image>0)
    return labels

def cluster_behaviors(dataset : VideosetDataFrame, 
                      feature_cols : list, 
                      N_rows : int = 200000, 
                      use_morlet : bool = False, 
                      use_umap :bool = True, 
                      n_pts : int = 300, 
                      bandwidth : float = 0.5,
                      **kwargs) -> tuple:
    """Cluster behaviors based on dimensionality reduction, kernel density estimation, and watershed clustering.
    
    **Note that this will modify the dataset dataframe in place.**

    The following columns are added to dataset: 
        'embedding_index_[0/1]': the coordinates of each embedding coordinate in the returned density matrix
        'unsup_behavior_label': the Watershed transform label for that row, based on its embedding coordinates. Rows whose embedding coordinate has no watershed cluster, or which fall outside the domain have value -1.

    Args:
        dataset: the VideosetDataFrame with the features of interest
        feature_cols: list of column names to perform the clustering on
        N_rows: number of rows to perform the embedding on. If 'None', then all rows are used.
        use_morlet: Apply Morlet wavelet transform to the feature cols before computing the embedding
        use_umap: If True will use UMAP dimensionality reduction, if False will use TSNE
        n_pts: dimension of grid the kernel density estimate is evaluated on. 
        bandwidth: Gaussian kernel bandwidth for kernel estimate
        **kwargs: All other keyword parameters are sent to dimensionality reduction call (either TSNE or UMAP)

    Returns:
        A tuple with components:
            dens_matrix: the (n_pts x n_pts) numpy array with the density estimate of the 2D embedding
            labels: numpy array with same dimensions are dens_matrix, but with values the watershed cluster IDs
            embedding_extent: the coordinates in embedding space that dens_matrix is approximating the density over
    """

    #If running interactively, can set default params here
    # N_rows = 200000
    # subsample = True 
    # use_morlet = False
    # use_umap = True
    # n_pts = 300
    # bandwidth = 0.5
    # kwargs = {}

    data = dataset.data[feature_cols]

    if use_morlet:
        print("Computing morlet transform of features")
        morlet = compute_morlet(data)
        morlet = morlet.reshape((morlet.shape[0]*morlet.shape[1], -1))
        morlet = morlet.T
        embedding_data = StandardScaler().fit_transform(morlet)
    else:
        embedding_data = StandardScaler().fit_transform(data)

    if N_rows is not None:
        random_indices = np.random.choice(embedding_data.shape[0], min(N_rows, embedding_data.shape[0]), replace = False)
        embedding_data_fit = embedding_data[random_indices, :]
    else:
        embedding_data_fit = embedding_data

    print("Performing dimensionality reduction")
    if use_umap:
        reducer = umap.UMAP(**kwargs)
        reducer.fit(embedding_data_fit)
        embedding = reducer.transform(embedding_data)
    else:
        if 'perplexity' not in kwargs:
            kwargs['perplexity'] = 50
        if 'init' not in kwargs:
            kwargs['init'] = 'pca'
        embedding = TSNE(n_components=2, **kwargs).fit_transform(embedding_data)

    dataset.data[['embedding_0', 'embedding_1']] = embedding

    xmin = np.quantile(dataset.data['embedding_0'], 0.01)
    xmax = np.quantile(dataset.data['embedding_0'], 0.99)
    ymin = np.quantile(dataset.data['embedding_1'], 0.01)
    ymax = np.quantile(dataset.data['embedding_1'], 0.99)

    xrange = (xmax - xmin)
    yrange = (ymax - ymin)
    xmin -= 0.1*xrange 
    xmax += 0.1*xrange
    ymin -= 0.1*yrange
    ymax += 0.1*yrange

    embedding_extent = (xmin, xmax, ymin, ymax)

    print("Computing density estimate of embedding")
    dens_matrix = compute_density(dataset, embedding_extent, bandwidth = bandwidth, n_pts = n_pts)
    print("Performing watershed transform of density")
    labels = compute_watershed(dens_matrix, positive_only = False, cutoff = 0)

    dataset.data['embedding_index_0'] = dataset.data['embedding_0'].apply(lambda x: int(max(0, min(n_pts-1, n_pts*(x-xmin)/(xmax-xmin)))))
    dataset.data['embedding_index_1'] = dataset.data['embedding_1'].apply(lambda x: int(max(0, min(n_pts-1, n_pts*(x-ymin)/(ymax-ymin)))))
    dataset.data['unsup_behavior_label'] = dataset.data[['embedding_index_0', 'embedding_index_1']].apply(lambda x: labels[x[1],x[0]], axis = 1)
    dataset.data.loc[dataset.data['unsup_behavior_label'] == 0, 'unsup_behavior_label'] = -1

    return (dens_matrix, labels, embedding_extent)


""" Loading and saving tracking and behavior annotation files """
import pandas as pd
import pickle
import numpy as np
from itertools import product
from joblib import dump, load
import os
from glob import glob
import warnings

XY_IDS = ['x', 'y']
XYLIKELIHOOD_IDS = ['x', 'y', 'likelihood']

def uniquifier(seq):
    """Return a sequence (e.g. list) with unique elements only, but maintaining original list order"""
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]

def _list_replace(ls, renamer):
    """Replace elements in a list according to provided dictionary"""
    for i, word in enumerate(ls):
        if word in renamer.keys():
            ls[i] = renamer[word]
    return ls

def save_sklearn_model(model, fn_out):
    """Save sklearn model to file
    
    Args:
        model: sklearn model to save
        fn_out: filename to save to
    """
    dump(model, fn_out) 

def load_sklearn_model(fn_in):
    """Load sklearn model from file
    
    Args:
        fn_in: filename to load from
    
    Returns:
        the loaded sklearn model
    """
    model = load(fn_in)
    return model 

def read_DLC_tracks(fn_in : str, 
                    part_renamer : dict = None, 
                    animal_renamer : dict = None,
                    read_likelihoods : bool = True) -> tuple:
    """Read in tracks from DLC.

    Args:
        fn_in: csv file that has DLC tracks
        part_renamer: dictionary to rename body parts, if needed 
        animal_renamer: dictionary to rename animals, if needed
        read_likelihoods: default True. Whether to attach DLC likelihoods to table

    Returns:
        Pandas DataFrame with (n_animals*2*n_body_parts) columns plus with filename and frame, 
            List of body parts,
            List of animals,
            Columns names for DLC tracks (excluding likelihoods, if read in),
            Scorer
    """

    df = pd.read_csv(fn_in, header = [0,1,2,3], index_col = 0)

    if 'individuals' in df.columns.names:
        df.columns = df.columns.set_names(['scorer', 'individuals', 'bodyparts', 'coords'])
        multi_animal = True
    else:
        df = pd.read_csv(fn_in, header = [0,1,2], index_col = 0)
        df.columns = df.columns.set_names(['scorer', 'bodyparts', 'coords'])
        multi_animal = False

    scorer = df.columns.get_level_values(0)[0]

    cols = list(df.columns)

    if multi_animal:
        animals = uniquifier([i[1] for i in cols])
        body_parts = uniquifier([i[2] for i in cols])
    else:
        animals = ['ind1']
        body_parts = uniquifier([i[1] for i in cols])

    n_body_parts = len(body_parts)
    n_animals = len(animals)

    dlc_tracks = np.array(df)
    n_rows = dlc_tracks.shape[0]
    selected_cols = [[3*i, 3*i+1] for i in range(n_body_parts*n_animals)]
    #This line flattens the array here... may be clearer using numpy reshape?
    selected_cols = [j for i in selected_cols for j in i]

    prob_cols = [3*i+2 for i in range(n_body_parts*n_animals)]
    dlc_probs = dlc_tracks[:,prob_cols]
    dlc_tracks = dlc_tracks[:,selected_cols]

    #Put in shape:
    # (frame, animal, x/y coord, body part)
    dlc_tracks = dlc_tracks.reshape((n_rows, n_animals, n_body_parts, 2))
    dlc_tracks = dlc_tracks.transpose([0, 1, 3, 2])

    dlc_probs = dlc_probs.reshape((n_rows, n_animals, n_body_parts, 1))
    dlc_probs = dlc_probs.transpose([0, 3, 1, 2])

    #If we're going to rename items in the list, do it here
    if part_renamer:
        body_parts = _list_replace(body_parts, part_renamer)

    if animal_renamer:
        animals = _list_replace(animals, animal_renamer)

    colnames = ['_'.join(a) for a in product(animals, XY_IDS, body_parts)]
    prob_colnames = ['_'.join(a) for a in product(['likelihood'], animals, body_parts)]

    dlc_probs = dlc_probs.reshape((n_rows, -1))
    final_probs = pd.DataFrame(dlc_probs, columns = prob_colnames)

    dlc_tracks = dlc_tracks.reshape((n_rows, -1))
    final_df = pd.DataFrame(dlc_tracks, columns = colnames)

    if read_likelihoods:
        final_df = pd.concat([final_df, final_probs], axis = 1)

    final_df['filename'] = fn_in
    final_df['frame'] = final_df.index.copy()

    return final_df, body_parts, animals, colnames, scorer

def rename_df_cols(df : pd.DataFrame, renamer : dict) -> pd.DataFrame:
    """Rename dataframe columns 
    
    Args:
        df: Pandas dataframe whose columns to rename
        renamer: dictionary whose key:value pairs define the substitutions to make

    Returns:
        The dataframe with renamed columns.
    """
    return df.rename(columns = renamer)

def save_DLC_tracks_h5(df : pd.DataFrame, fn_out : str) -> None:
    """Save DLC tracks in h5 format.
    
    Args:
        df: Pandas dataframe to save
        fn_out: Where to save the dataframe
    """
    df.to_hdf(fn_out, "df_with_missing", format = 'table', mode="w")

def load_data(fn : str):
    """Load an object from a pickle file
    
    Args:
        fn: The filename

    Returns:
        The pickled object.
    """
    with open(fn, 'rb') as handle:
        object = pickle.load(handle)
    return object 

def _make_sample_dataframe(fn_out = 'sample_dataframe.pkl'): # pragma: no cover
    from behaveml import VideosetDataFrame, clone_metadata

    cur_dir = os.path.dirname(os.path.abspath(__file__))
    tracking_files = sorted(glob(cur_dir + '/data/dlc/*.csv'))
    boris_files = sorted(glob(cur_dir + '/data/boris/*.csv'))
    frame_width = None               # (float) length of entire horizontal shot
    frame_width_units = None         # (str) units frame_width is given in
    fps = 30                         # (int) frames per second
    resolution = (1200, 1600)        # (tuple) HxW in pixels
    metadata = clone_metadata(tracking_files, 
                          label_files = boris_files, 
                          frame_width = frame_width, 
                          fps = fps, 
                          frame_width_units = frame_width_units, 
                          resolution = resolution)

    dataset = VideosetDataFrame(metadata)
    path_out = os.path.join(cur_dir, 'data', fn_out)
    to_save = {'dataset': dataset, 'metadata': metadata}
    with open(path_out, 'wb') as handle:
        pickle.dump(to_save, handle, protocol=pickle.HIGHEST_PROTOCOL)

def get_sample_data_paths():
    """Get path to sample data files provided with package. 
    
    Returns:
        (tuple) list of DLC tracking file, list of boris annotation files
    """
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    tracking_files = sorted(glob(cur_dir + '/data/dlc/*.csv'))
    boris_files = sorted(glob(cur_dir + '/data/boris/*.csv'))
    return tracking_files, boris_files

def get_sample_data():
    """Load a sample dataset of 5 mice social interaction videos. Each video is approx. 5 minutes in duration
    
    Returns:
        (VideosetDataFrame) Data frame with the corresponding tracking and behavior annotation files
    """

    cur_dir = os.path.dirname(os.path.abspath(__file__))
    path_in = os.path.join(cur_dir, 'data', 'sample_dataframe.pkl')
    with open(path_in, 'rb') as handle:
        b = pickle.load(handle)
    return b['dataset']

def read_boris_annotation(fn_in : str, fps : int, duration : float, behav_labels : dict = None) -> tuple:
    """Read behavior annotation from BORIS exported csv file. 

    This will import behavior types specified (or all types, if behavior_list is None) and assign a numerical label to each. Overlapping annotations (those occurring simulataneously) are not supported. Any time the video is annotated as being in multiple states, the last state will be the one labeled.
    
    Args:
        fn_in: The filename with BORIS behavior annotations to load
        fps: The frames per second of the video
        duration: The duration of the video in seconds
        behav_labels: If provided, only import behaviors with these names. Default = None = import everything. 
    
    Returns:
        A numpy array which indicates, for all frames, which behavior is occuring. 0 = no behavior, 1 and above are the labels of the behaviors.
        A dictionary with keys the numerical labels and values the names of the behaviors. 
    """

    boris_labels = pd.read_csv(fn_in, skiprows = 15)
    if len(boris_labels) == 0:
        print("No data found in BORIS file,", fn_in)
        return np.array([]), {}
    fps_boris = int(boris_labels['FPS'][0])
    duration_boris = boris_labels['Total length'][0]
    if fps_boris != fps:
        warnings.warn(f"Warning: BORIS FPS is {fps_boris} but video is {fps} frames per second. Are the DLC and BORIS files from the same video?")
    if not np.isclose(duration_boris, duration, rtol = 0.01, atol = 0.01):
        warnings.warn(f"Warning: BORIS duration is {duration_boris} but video is {duration} seconds. Are the DLC and BORIS files from the same video?")

    n_bins = int(duration*fps)
    ground_truth = np.zeros(n_bins)

    if behav_labels is None:
        behaviors = boris_labels['Behavior'].unique()
        behav_labels = {i+1:k for i,k in enumerate(behaviors)}

    for behav_idx, behavior in behav_labels.items():
        labels = boris_labels[boris_labels['Behavior'] == behavior]
        starts = labels.loc[labels['Status'] == 'START', 'Time']
        ends = labels.loc[labels['Status'] == 'STOP', 'Time']
        if len(ends) > len(starts)+1:
            raise ValueError(f"Too many {behavior} behaviors started and not stopped.")
        elif len(ends) == len(starts) + 1:
            ends.append(duration)
        for start, end in zip(starts, ends):
            ground_truth[int(start*fps):int(end*fps)] = behav_idx

    return ground_truth, behav_labels

def create_behavior_labels(boris_files):
    """Create behavior labels from BORIS exported csv files.
    
    Args:
        boris_files: List of BORIS exported csv files
        
    Returns:
        A dictionary with keys the numerical labels and values the names of the behaviors.
    """
    behaviors = set()
    for fn in boris_files:
        boris_labels = pd.read_csv(fn, skiprows = 15)
        behaviors = behaviors | set(boris_labels['Behavior'].unique())
    behavior_labels = {i+1:k for i,k in enumerate(behaviors)}
    return behavior_labels


import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import cm
import os
import time
import numpy as np
from glob import glob
from behaveml.config import global_config
from behaveml import VideosetDataFrame
import pandas as pd

def plot_embedding(dataset : VideosetDataFrame, 
                   col_names : list  = ['embedding_0', 'embedding_1'],
                   color_col : str = None, 
                   figsize : tuple = (10,10),
                   **kwargs) -> tuple:
    """Scatterplot of a 2D TSNE or UMAP embedding from the dataset.
    
    Args:
        dataset: data
        col_names: list of column names to use for the x and y axes
        color_col: if provided, a column that will be used to color the points in the scatter plot
        figsize: tuple with the dimensions of the plot (in inches)
        kwargs: All other keyword pairs are sent to Matplotlib's scatter function

    Returns:
        tuple (fig, axes). The Figure and Axes objects. 
    """

    if color_col is not None:
        c = dataset.data[color_col]
    else:
        c = None

    col1, col2 = col_names
    fig, axes = plt.subplots(1,1, figsize = figsize)
    axes.scatter(x = dataset.data[col1], y = dataset.data[col2], s = 1, c = c, **kwargs)
    axes.set_xlabel('Embedding dim 1')
    axes.set_ylabel('Embedding dim 2')
    return fig, axes

class MplColorHelper:

    def __init__(self, cmap_name, start_val, stop_val):
        self.cmap_name = cmap_name
        self.cmap = plt.get_cmap(cmap_name)
        self.norm = mpl.colors.Normalize(vmin=start_val, vmax=stop_val)
        self.scalarMap = cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

    def get_rgb(self, val):
        return self.scalarMap.to_rgba(val)
    

def plot_unsupervised_results(dataset : VideosetDataFrame, 
                              cluster_results : tuple, 
                              col_names : list = ['embedding_0', 'embedding_1'],
                              figsize : tuple = (15,4), 
                              **kwargs):
    """Set of plots for unsupervised behavior clustering results
    
    Args:
        dataset: data
        cluster_results: tuple output by 'cluster_behaviors'
        col_names: list of column names to use for the x and y axes
        figsize: tuple with the plot dimensions, in inches
        kwargs: all other keyword pairs are sent to Matplotlib's scatter function

    Returns:
        tuple (fig, axes). The Figure and Axes objects. 
    """    
    dens_matrix, labels, extent = cluster_results

    fig, axes = plt.subplots(1, 3, figsize = figsize)
    col1, col2 = col_names
    axes[0].scatter(x = dataset.data[col1], y = dataset.data[col2], s = 1, **kwargs)
    axes[0].set_xlabel('Embedding dim 1')
    axes[0].set_ylabel('Embedding dim 2')
    axes[0].set_title('Embedding')
    rectangle = plt.Rectangle((extent[0],extent[2]), extent[1]-extent[0], extent[3]-extent[2], fill=False,ec="red")
    axes[0].add_patch(rectangle)
    axes[1].imshow(dens_matrix, origin = 'lower')
    axes[1].set_title('Density estimate')
    axes[1].axis('off')
    axes[2].imshow(labels, origin = 'lower')
    axes[2].axis('off')
    axes[2].set_title('Watershed clustering')

    #For each label, find the mean location
    all_labels = np.unique(labels)
    all_labels = all_labels[all_labels != 0]
    colhelper = MplColorHelper('viridis', 0, max(all_labels))

    for lb in all_labels:
        locs = np.where(labels == lb)
        mean_y = np.mean(locs[0])
        mean_x = np.mean(locs[1])
        lab_col = colhelper.get_rgb(lb)
        lab_bright = lab_col[0]*0.299 + lab_col[1]*0.587 + lab_col[2]*0.114 > (120/255)
        text_col = 'black' if lab_bright else 'white'
        axes[2].text(mean_x-8, mean_y-4, str(lb), c = text_col)

    return fig, axes

def plot_ethogram(dataset : VideosetDataFrame, 
                  vid_key : str, 
                  query_label : str = 'unsup_behavior_label', 
                  frame_limit : int = 4000, 
                  figsize : tuple = (16,2)) -> tuple:
    """Simple ethogram of one video, up to a certain frame number.

    Args:
        dataset: 
        vid_key: key (in dataset.metadata) pointing to the video to make ethogram for
        query_label: the column containing the behavior labels to plot
        frame_limit: only make the ethogram for frames between [0, frame_limit]
        figsize: tuple with figure size (in inches)

    Returns:
        tuple (fig, axes). The Figure and Axes objects
    """
    fig, ax = plt.subplots(1,1,figsize = figsize)
    plot_data = dataset.data.loc[dataset.data['filename'] == vid_key, query_label][:frame_limit].to_numpy()
    b = np.zeros((plot_data.size, plot_data.max()+1))
    b[np.arange(plot_data.size),plot_data] = plot_data
    plt.imshow(b.T, aspect = 'auto', origin = 'lower', interpolation = 'none', alpha = (b.T > 0).astype(float))
    plt.axis('off')
    plt.tight_layout(pad = 0)
    plt.xlim([0, frame_limit])
    return fig, ax

#TODO
# Bug with trimming the jpg here. 
def create_ethogram_video(dataset : VideosetDataFrame, 
                          vid_key : str, 
                          query_label : str, 
                          out_file : str, 
                          frame_limit : int = 4000, 
                          im_dim : float = 16, 
                          min_frames : int = 3) -> None:
    """Overlay ethogram on top of source video with ffmpeg

    Args:
        dataset: source dataset
        vid_key: the key (in dataset.metadata) pointing to the video to make ethogram for. metadata must have field 'video_files' that points to the source video location
        query_label: the column containing the behavior labels to plot
        out_file: output path for created video
        frame_limit: only make the ethogram/video for frames [0, frame_limit]
        in_dim: x dimension (in inches) of ethogram
        min_frames: any behaviors occurring for less than this number of frames are not labeled

    Returns:
        None
    """
    vid_file = dataset.metadata[vid_key]['video_files']
    fps = dataset.metadata[vid_key]['fps']
    time_limit = frame_limit/fps
    fig, _ = plt.subplots(1,1,figsize = (im_dim,2))
    plot_data = dataset.data.loc[dataset.data['filename'] == vid_key, query_label][:frame_limit].to_numpy()

    #Create behavior_times
    behavior_times = []
    cur_behav = -2
    start_idx = 0
    for idx, lab in enumerate(plot_data):
        if cur_behav != lab:
            if cur_behav != -2:
                if idx - start_idx > min_frames:
                    behavior_times.append((start_idx/fps, (idx-1)/fps, cur_behav))
            start_idx = idx
        cur_behav = lab
    behavior_times.append((start_idx/fps, (idx-1)/fps, cur_behav))

    b = np.zeros((plot_data.size, plot_data.max()+1))
    b[np.arange(plot_data.size),plot_data] = plot_data
    plt.ioff()
    plt.imshow(b.T, aspect = 'auto', origin = 'lower', interpolation = 'none', alpha = (b.T > 0).astype(float))
    plt.axis('off')
    plt.tight_layout(pad = 0)
    plt.xlim([0, frame_limit])
    os.makedirs('./tmp/', exist_ok = True)
    fn_out = f'./tmp/{query_label}.jpg'
    trimmed_fn = fn_out.replace('.jpg', '_trimmed.jpg')
    dpi = 2000/im_dim
    fig.savefig(fn_out, dpi = dpi)
    plt.close()
    plt.ioff()
    trim_cmd = f'convert {fn_out} -fuzz 7% -trim -resize 1600x1600 {trimmed_fn}'
    os.system(trim_cmd)
    vid_path = vid_file
    #Combine query label with movie    
    start_time_str = time.strftime('%H:%M:%S', time.gmtime(time_limit))
    text_filter = f"drawtext=text='|':fontcolor=green:fontsize=60:y=1:x='-10+(mod(round((w+5)*t/{time_limit}),w+5))',drawtext=text='|':fontcolor=green:fontsize=60:y=20:x='-10+(mod(round((w+5)*t/{time_limit}),(w+5)))'"
    for str_time, end_time, behav_label in behavior_times:
        behav_text = f",drawtext=text='{behav_label}':fontcolor=black:fontsize=30:y=10:x=10:enable='between(t,{str_time},{end_time})'"
        text_filter += behav_text

    ffmpeg_cmd = f'''ffmpeg -y -hide_banner -loglevel error -i {vid_path} -i {trimmed_fn} \
    -filter_complex "[0:v][1:v]overlay=0:0,{text_filter}" \
    -t {start_time_str} \
    -threads 8 -q:v 3 {out_file}'''
    os.system(ffmpeg_cmd)

def create_sample_videos(dataset : VideosetDataFrame, 
                         video_dir : str, 
                         out_dir : str, 
                         query_col : str = 'unsup_behavior_label', 
                         N_sample_rows : int = 16, 
                         window_size : int = 2, 
                         fps : float = 30,
                         N_supersample_rows : int = 1000) -> None:
    """Create a sample of videos displaying the labeled behaviors using ffmpeg. 

    For each behavior label, randomly choose frames from the entire dataset and extract short clips from source videos based around those points. Tries to select frames where the labeled behavior is exhibited in many frames of the clip.

    Args:
        dataset: source dataset
        video_dir: location of source video files
        out_dir: base output directory to save videos. Videos are saved in the form: [out_dir]/[behavior_label]/[video_name]_[time in seconds].avi
        query_label: the column containing the behavior labels to extract clips for. Each unique value in this column is treated as a separate behavior
        N_sample_rows: number of clips to extract per behavior
        window_size: amount of video to extract on either side of the sampled frame, in seconds
        fps: frames per second of videos
        N_supersample_rows: this many rows are randomly sampled for each behavior label, and the top N_sample_rows are returned (in terms of number of adjacent frames also exhibiting that behavior). Shouldn't need to play with this.

    Returns:
        None
    """
    labels = dataset.data[query_col].unique()
    labels = labels[labels >= 0]
    #all_labels = np.unique(labels)

    def get_window_size(label_idx, sample_row, max_size = 500):
        s_m = 0
        for idx in range(max_size):
            try:
                if dataset.data.loc[sample_row-idx, query_col] == label_idx:
                    s_m += 1
                else:
                    break
            except:
                break
                    
        s_p = 0
        for idx in range(max_size):
            try:
                if dataset.data.loc[sample_row+idx, query_col] == label_idx:
                    s_p += 1
                else:
                    break
            except:
                break
        return sample_row-s_m, s_p+sample_row, sample_row, s_m+s_p


    for label_idx in labels:

        print(f"Making sample videos for behavior label {label_idx}")
        label_indices = dataset.data[query_col] == label_idx
        if sum(label_indices) == 0: continue

        ## Pull out some sample frames from each video for this behavior
        behavior_rows = dataset.data[dataset.data[query_col] == label_idx]
        random_sample_indices = np.random.choice(behavior_rows.index, min(len(behavior_rows.index), N_supersample_rows), replace = False)
        window_sizes = [get_window_size(label_idx, i) \
                        for i in random_sample_indices]
        window_sizes = sorted(window_sizes, key = lambda x: x[3], reverse = True)
        window_sizes = pd.DataFrame(window_sizes, columns = ['start', 'end', 'pt', 'len']).drop_duplicates(subset = ['start', 'end'])

        random_sample_indices = window_sizes['pt'].to_numpy()[:N_sample_rows]

        behavior_rows_sample = behavior_rows.loc[random_sample_indices].reset_index(drop = True)
        #For each filename in this list of samples, extract a part of that video with ffmpeg
        filenames = behavior_rows_sample.filename.unique()
        video_files = [os.path.basename(p).split('DLC')[0]+'.avi' for p in filenames]

        out_dir_vid = os.path.join(out_dir, f'behavior_label_{label_idx}')
        os.makedirs(out_dir_vid, exist_ok = True)


        for r_idx, (vid_file, fn) in enumerate(zip(video_files, filenames)):
            behave_rows_sample_vid = behavior_rows_sample[behavior_rows_sample['filename'] == fn]
            vid_name = os.path.basename(vid_file).split('.')[0]
            for idx in range(len(behave_rows_sample_vid)):

                sampl_window = window_sizes.iloc[r_idx,:].to_numpy()
                str_time = window_size + (sampl_window[0] - sampl_window[2])/fps
                end_time = window_size + (sampl_window[1] - sampl_window[2])/fps
                #str_time = 0
                #end_time = 2

                frame_number = behave_rows_sample_vid.reset_index().loc[idx, 'frame']
                behavior_time = int(frame_number/fps)
                out_file = os.path.join(out_dir_vid, f'{vid_name}_second_{behavior_time}.avi')
                start_time = max(0, behavior_time - window_size)
                start_time_str = time.strftime('%H:%M:%S', time.gmtime(start_time))
                #TODO
                # Get the times right on this line...
                text_filter = f"drawtext=text='{label_idx} active':fontcolor=red:fontsize=100:y=10:x=10:box=1:boxcolor=white:enable='between(t,{str_time},{end_time})'"
                ffmpeg_cmd = f'''ffmpeg -y -hide_banner -loglevel error -ss {start_time_str} -i {os.path.join(video_dir, vid_file)} -t 00:00:{2*window_size} \
                    -filter_complex "{text_filter}" \
                    -threads 4 {out_file}'''
                os.system(ffmpeg_cmd)

def create_mosaic_video(vid_dir : str, 
                        output_file : str, 
                        ndim : tuple = (1600, 1200)) -> None:
    """Take a set of video clips and turn them into a mosaic using ffmpeg 
    
    16 videos are tiled.

    Args:
        vid_dir: source directory with videos in it
        output_file: output video path
        ndim: tuple with the output video dimensions, in pixels
    
    Returns:
        None    
    """
    max_mosaic_vids = global_config['create_mosaic_video__max_mosaic_vids']
    mosaic_vid_files = glob(vid_dir)[:max_mosaic_vids]

    block_size_x = ndim[0]/np.sqrt(max_mosaic_vids)
    block_size_y = ndim[1]/np.sqrt(max_mosaic_vids)

    mosaic_cmd = f'''ffmpeg -y -hide_banner -loglevel error \
    {' '.join([f'-i {f}' for f in mosaic_vid_files])} \
    -filter_complex " \
        nullsrc=size={'x'.join([str(i) for i in ndim])} [base]; \
        [0:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [upper1]; \
        [1:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [upper2]; \
        [2:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [upper3]; \
        [3:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [upper4]; \
        [4:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [uppermid1]; \
        [5:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [uppermid2]; \
        [6:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [uppermid3]; \
        [7:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [uppermid4]; \
        [8:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lowermid1]; \
        [9:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lowermid2]; \
        [10:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lowermid3]; \
        [11:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lowermid4]; \
        [12:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lower1]; \
        [13:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lower2]; \
        [14:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lower3]; \
        [15:v] setpts=PTS-STARTPTS, scale={block_size_x}x{block_size_y} [lower4]; \
        [base][upper1] overlay=shortest=1 [tmp1]; \
        [tmp1][upper2] overlay=shortest=1:x={block_size_x} [tmp2]; \
        [tmp2][upper3] overlay=shortest=1:x={2*block_size_x} [tmp3]; \
        [tmp3][upper4] overlay=shortest=1:x={3*block_size_x} [tmp4];\
        [tmp4][uppermid1] overlay=shortest=1:y={block_size_y} [tmp5]; \
        [tmp5][uppermid2] overlay=shortest=1:x={block_size_x}:y={block_size_y} [tmp6]; \
        [tmp6][uppermid3] overlay=shortest=1:x={2*block_size_x}:y={block_size_y} [tmp7]; \
        [tmp7][uppermid4] overlay=shortest=1:x={3*block_size_x}:y={block_size_y} [tmp8];\
        [tmp8][lowermid1] overlay=shortest=1:y={2*block_size_y} [tmp9]; \
        [tmp9][lowermid2] overlay=shortest=1:x={block_size_x}:y={2*block_size_y} [tmp10]; \
        [tmp10][lowermid3] overlay=shortest=1:x={2*block_size_x}:y={2*block_size_y} [tmp11]; \
        [tmp11][lowermid4] overlay=shortest=1:x={3*block_size_x}:y={2*block_size_y} [tmp12];\
        [tmp12][lower1] overlay=shortest=1:y={3*block_size_y} [tmp13]; \
        [tmp13][lower2] overlay=shortest=1:x={block_size_x}:y={3*block_size_y} [tmp14]; \
        [tmp14][lower3] overlay=shortest=1:x={2*block_size_x}:y={3*block_size_y} [tmp15]; \
        [tmp15][lower4] overlay=shortest=1:x={3*block_size_x}:y={3*block_size_y} \
    " -c:v libx264 {output_file}'''
    os.system(mosaic_cmd)


import pandas as pd 
from behaveml.dl.feature_engineering import make_features_mars_distr, make_features_social, \
                                            make_features_distances, make_features_velocities, \
                                            make_features_mars_reduced

def compute_mars_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:
    features_df = make_features_mars_distr(df[raw_col_names], animal_setup) 
    return features_df

def compute_distance_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:
    features_df = make_features_distances(df[raw_col_names], animal_setup) 
    return features_df

def compute_mars_reduced_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:
    features_df = make_features_mars_reduced(df[raw_col_names], animal_setup) 
    return features_df

def compute_social_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:
    features_df = make_features_social(df[raw_col_names], animal_setup) 
    return features_df

def compute_velocity_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:
    features_df = make_features_velocities(df[raw_col_names], animal_setup) 
    return features_df



""" Functions to take pose tracks and compute a set of features from them """

import pandas as pd
import numpy as np

# TODO
# * More generic feature creation functions, for general models, not just MARS featureset
#   - [ ] Intra-animal angles (of all body parts)

############################

def _diff_within_group(df, sort_key, diff_col, **kwargs):
    return df.groupby(sort_key)[diff_col].transform(lambda x: x.diff(**kwargs)) 

def compute_centerofmass_interanimal_distances(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']

    features_df = df.copy()

    for animal_id in mouse_ids:
        fxs = ['_'.join([animal_id, 'x', bp]) for bp in bodypart_ids]
        fys = ['_'.join([animal_id, 'y', bp]) for bp in bodypart_ids]
        fx_new = '_'.join([animal_id, 'COM_x'])
        fy_new = '_'.join([animal_id, 'COM_y'])
        features_df[fx_new] = features_df[fxs].sum(axis = 1) / len(bodypart_ids)
        features_df[fy_new] = features_df[fys].sum(axis = 1) / len(bodypart_ids)

    orig_cols = features_df.columns

    for i, animal_i in enumerate(mouse_ids):
        fx_i = '_'.join([animal_i, 'COM_x'])
        fy_i = '_'.join([animal_i, 'COM_y'])
        for j, animal_j in enumerate(mouse_ids[:i]):
            fx_j = '_'.join([animal_j, 'COM_x'])
            fy_j = '_'.join([animal_j, 'COM_y'])
            f_new = '_'.join([animal_i, animal_j, 'COM_distance'])
            features_df[f_new] = np.sqrt((features_df[fx_i] - features_df[fx_j])**2 \
                                         + (features_df[fy_i] - features_df[fy_j])**2)

    features_df = features_df.drop(columns = orig_cols)
    return features_df

def compute_centerofmass_interanimal_speed(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, n_shifts = 5, **kwargs) -> pd.DataFrame:

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']

    features_df = df.copy()

    dt = features_df['time'].diff(periods = n_shifts)

    for animal_id in mouse_ids:
        fxs = ['_'.join([animal_id, 'x', bp]) for bp in bodypart_ids]
        fys = ['_'.join([animal_id, 'y', bp]) for bp in bodypart_ids]
        fx_new = '_'.join([animal_id, 'COM_x'])
        fy_new = '_'.join([animal_id, 'COM_y'])
        features_df[fx_new] = features_df[fxs].sum(axis = 1) / len(bodypart_ids)
        features_df[fy_new] = features_df[fys].sum(axis = 1) / len(bodypart_ids)
        features_df[fx_new] = _diff_within_group(features_df, 'filename', fx_new, periods = n_shifts)/dt
        features_df[fy_new] = _diff_within_group(features_df, 'filename', fy_new, periods = n_shifts)/dt

    orig_cols = features_df.columns

    for i, animal_i in enumerate(mouse_ids):
        fx_i = '_'.join([animal_i, 'COM_x'])
        fy_i = '_'.join([animal_i, 'COM_y'])
        for _, animal_j in enumerate(mouse_ids[:i]):
            fx_j = '_'.join([animal_j, 'COM_x'])
            fy_j = '_'.join([animal_j, 'COM_y'])
            f_new = '_'.join([animal_i, animal_j, 'COM_speed'])

            vx_i = _diff_within_group(features_df, 'filename', fx_i, periods = n_shifts)/dt
            vy_i = _diff_within_group(features_df, 'filename', fy_i, periods = n_shifts)/dt
            vx_j = _diff_within_group(features_df, 'filename', fx_j, periods = n_shifts)/dt
            vy_j = _diff_within_group(features_df, 'filename', fy_j, periods = n_shifts)/dt

            features_df[f_new] = np.sqrt((vx_i - vx_j)**2 + (vy_i - vy_j)**2)

    features_df = features_df.drop(columns = orig_cols)
    return features_df

def compute_centerofmass(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, bodyparts : list = [], **kwargs) -> pd.DataFrame:

    if len(bodyparts) == 0:
        bodypart_ids = animal_setup['bodypart_ids']
    else:
        bodypart_ids = [v for v in bodyparts if v in animal_setup['bodypart_ids']]
    if len(bodypart_ids) == 0:
        raise ValueError('No listed bodyparts found in animal_setup')

    mouse_ids = animal_setup['mouse_ids']

    features_df = df.copy()
    orig_cols = df.columns

    for animal_id in mouse_ids:
        fxs = ['_'.join([animal_id, 'x', bp]) for bp in bodypart_ids]
        fys = ['_'.join([animal_id, 'y', bp]) for bp in bodypart_ids]
        fx_new = '_'.join([animal_id, 'COM_x'])
        fy_new = '_'.join([animal_id, 'COM_y'])
        features_df[fx_new] = features_df[fxs].sum(axis = 1) / len(bodypart_ids)
        features_df[fy_new] = features_df[fys].sum(axis = 1) / len(bodypart_ids)

    features_df = features_df.drop(columns = orig_cols)
    return features_df

# def compute_centerofmass(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:

#     bodypart_ids = animal_setup['bodypart_ids']
#     mouse_ids = animal_setup['mouse_ids']

#     features_df = df.copy()
#     orig_cols = df.columns

#     for animal_id in mouse_ids:
#         fxs = ['_'.join([animal_id, 'x', bp]) for bp in bodypart_ids]
#         fys = ['_'.join([animal_id, 'y', bp]) for bp in bodypart_ids]
#         fx_new = '_'.join([animal_id, 'COM_x'])
#         fy_new = '_'.join([animal_id, 'COM_y'])
#         features_df[fx_new] = features_df[fxs].sum(axis = 1) / len(bodypart_ids)
#         features_df[fy_new] = features_df[fys].sum(axis = 1) / len(bodypart_ids)

#     features_df = features_df.drop(columns = orig_cols)
#     return features_df

############################

def compute_centerofmass_velocity(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, n_shifts = 5, bodyparts : list = [], **kwargs) -> pd.DataFrame:

    if len(bodyparts) == 0:
        bodypart_ids = animal_setup['bodypart_ids']
    else:
        bodypart_ids = [v for v in bodyparts if v in animal_setup['bodypart_ids']]
    if len(bodypart_ids) == 0:
        raise ValueError('No listed bodyparts found in animal_setup')

    mouse_ids = animal_setup['mouse_ids']

    features_df = df.copy()
    orig_cols = df.columns

    dt = features_df['time'].diff(periods = n_shifts)

    for animal_id in mouse_ids:
        fxs = ['_'.join([animal_id, 'x', bp]) for bp in bodypart_ids]
        fys = ['_'.join([animal_id, 'y', bp]) for bp in bodypart_ids]
        fx_new = '_'.join([animal_id, 'COM_vel_x'])
        fy_new = '_'.join([animal_id, 'COM_vel_y'])
        features_df[fx_new] = features_df[fxs].sum(axis = 1) / len(bodypart_ids)
        features_df[fy_new] = features_df[fys].sum(axis = 1) / len(bodypart_ids)
        features_df[fx_new] = _diff_within_group(features_df, 'filename', fx_new, periods = n_shifts)/dt
        features_df[fy_new] = _diff_within_group(features_df, 'filename', fy_new, periods = n_shifts)/dt

    features_df = features_df.drop(columns = orig_cols)
    return features_df


def compute_speed_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, n_shifts = 5, **kwargs) -> pd.DataFrame:

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']

    features_df = df.copy()
    orig_cols = df.columns

    dt = features_df['time'].diff(periods = n_shifts)

    ##Make the distance features
    for i, bp1 in enumerate(bodypart_ids):
        for j, bp2 in enumerate(bodypart_ids):
            if i < j:
                for mouse_id in mouse_ids:
                    #We can compute the intra-mouse difference
                    f1x = '_'.join([mouse_id, 'x', bp1])
                    f2x = '_'.join([mouse_id, 'x', bp2])
                    f1y = '_'.join([mouse_id, 'y', bp1])
                    f2y = '_'.join([mouse_id, 'y', bp2])
                    f_new = '_'.join([mouse_id, 'speed', bp1, bp2])

                    vx1 = _diff_within_group(features_df, 'filename', f1x, periods = n_shifts)/dt
                    vy1 = _diff_within_group(features_df, 'filename', f1y, periods = n_shifts)/dt
                    vx2 = _diff_within_group(features_df, 'filename', f2x, periods = n_shifts)/dt
                    vy2 = _diff_within_group(features_df, 'filename', f2y, periods = n_shifts)/dt
                    features_df[f_new] = np.sqrt((vx1 - vx2)**2 + (vy1 - vy2)**2)

            #Inter-mouse difference
            for animal_i in range(len(mouse_ids)):
                for animal_j in range(animal_i):
                    f1x = '_'.join([mouse_ids[animal_i], 'x', bp1])
                    f2x = '_'.join([mouse_ids[animal_j], 'x', bp2])
                    f1y = '_'.join([mouse_ids[animal_i], 'y', bp1])
                    f2y = '_'.join([mouse_ids[animal_j], 'y', bp2])
                    f_new = '_'.join([f'M{animal_i}_M{animal_j}', 'speed', bp1, bp2])

                    vx1 = _diff_within_group(features_df, 'filename', f1x, periods = n_shifts)/dt
                    vy1 = _diff_within_group(features_df, 'filename', f1y, periods = n_shifts)/dt
                    vx2 = _diff_within_group(features_df, 'filename', f2x, periods = n_shifts)/dt
                    vy2 = _diff_within_group(features_df, 'filename', f2y, periods = n_shifts)/dt
                    features_df[f_new] = np.sqrt((vx1 - vx2)**2 + (vy1 - vy2)**2)

    #Remove base features
    features_df = features_df.drop(columns = orig_cols)

    return features_df

def compute_distance_features(df : pd.DataFrame, raw_col_names : list, animal_setup : dict, **kwargs) -> pd.DataFrame:

    bodypart_ids = animal_setup['bodypart_ids']
    mouse_ids = animal_setup['mouse_ids']

    features_df = df.copy()
    orig_cols = df.columns

    ##Make the distance features
    for i, bp1 in enumerate(bodypart_ids):
        for j, bp2 in enumerate(bodypart_ids):
            if i < j:
                for mouse_id in mouse_ids:
                    #We can compute the intra-mouse difference
                    f1x = '_'.join([mouse_id, 'x', bp1])
                    f2x = '_'.join([mouse_id, 'x', bp2])
                    f1y = '_'.join([mouse_id, 'y', bp1])
                    f2y = '_'.join([mouse_id, 'y', bp2])
                    f_new = '_'.join([mouse_id, 'distance', bp1, bp2])
                    features_df[f_new] = \
                        np.sqrt((features_df[f1x] - features_df[f2x])**2 + 
                                (features_df[f1y] - features_df[f2y])**2)

            #Inter-mouse difference
            for animal_i in range(len(mouse_ids)):
                for animal_j in range(animal_i):
                    f1x = '_'.join([mouse_ids[animal_i], 'x', bp1])
                    f2x = '_'.join([mouse_ids[animal_j], 'x', bp2])
                    f1y = '_'.join([mouse_ids[animal_i], 'y', bp1])
                    f2y = '_'.join([mouse_ids[animal_j], 'y', bp2])
                    f_new = '_'.join([f'M{animal_i}_M{animal_j}', 'distance', bp1, bp2])
                    features_df[f_new] = \
                                np.sqrt((features_df[f1x] - features_df[f2x])**2 + 
                                        (features_df[f1y] - features_df[f2y])**2)

    #Remove base features
    features_df = features_df.drop(columns = orig_cols)

    return features_df


""" Basic video tracking and behavior class that houses data. 

Basic object is the VideoSetDataFrame class.

## A note on unit conversions

For the unit rescaling, if the dlc/tracking file is already in desired units, either in physical distances, or pixels, then don't provide all of 'frame_width', 'resolution', and 'frame_width_units'. If you want to keep track of the units, you can add a 'units' key to the metadata. This could be 'pixels', or 'cm', as appropriate.

If the tracking is in pixels and you do want to rescale it to some physical distance, you should provide 'frame_width', 'frame_width_units' and 'resolution' for all videos. This ensures the entire dataset is using the same units. The package will use these values for each video to rescale the (presumed) pixel coordinates to physical coordinates. 

Resolution is a tuple (H,W) in pixels of the videos. 'frame_width' is the width of the image, in units 'frame_width_units'

When this is done, all coordinates are converted to 'mm'. The pair 'units':'mm' is added to the metadata dictionary for each video

If any of the provided parameters are provided, but are not the right format, or some values are missing, a warning is given and the rescaling is not performed.
"""

import pandas as pd
import pickle 
import os
import numpy as np
import re
from glob import glob
from sklearn.model_selection import PredefinedSplit

from behaveml.features import Features

from behaveml.io import read_DLC_tracks, read_boris_annotation, uniquifier, create_behavior_labels
from behaveml.utils import checkFFMPEG

from behaveml.config import global_config

#This converts everything to mm, or leaves them as pixels
UNIT_DICT = {'mm':1, 'cm':10, 'm':1000, 'in':25.4, 'ft':304.8,
             'millimeters':1, 'centimeters':10, 'meters':1000, 'inches':25.4, 'feet':304.8,
             'millimeter':1, 'centimeter':10, 'meter':1000, 'inch':25.4, 'foot':304.8,
             'millimetres':1, 'centimetres':10, 'metres':1000,
             'millimetre':1, 'centimetre':10, 'metre':1000}

class MLDataFrame(object): # pragma: no cover
    """
    DataFrame useful for interfacing between pandas and sklearn. Stores a data
    table and metadata dictionary. When feature columns, label columns and fold columns are specified
    then creates properties features, labels, folds and splitter that sklearn accepts for ML.
    """
    def __init__(self, data : pd.DataFrame,
                       metadata : dict = {},
                       fold_cols = None,
                       feature_cols = None,
                       label_cols = None):
        self.data = data 
        self.metadata = metadata
        self.fold_cols = fold_cols
        self.feature_cols = feature_cols
        self.label_cols = label_cols

    def add_data(self, new_data, col_names):
        if type(new_data) == pd.DataFrame:
            self.data[col_names] = new_data
        else:
            self.data[col_names] = pd.DataFrame(new_data, index=self.data.index)

    @property
    def features(self):
        if self.feature_cols is None:
            return None
        else:
            return self.data[self.feature_cols].to_numpy()

    @property
    def labels(self):
        if self.label_cols is None:
            return None
        else:
            return self.data[self.label_cols].to_numpy()

    @property
    def folds(self):
        if self.fold_cols is None:
            return None
        else:
            return self.data[self.fold_cols].to_numpy()

    @property
    def splitter(self):
        if self.fold_cols is None:
            return None
        else:
            return self._make_predefined_split(self.folds)

    def _make_predefined_split(self, folds):
        test_indices = np.sum(folds, axis=1) == 0
        test_fold = -1*(test_indices)
        test_fold[test_indices == False] = np.argmin(folds[test_indices == False], axis = 1)
        return PredefinedSplit(test_fold)

    def save(self, fn):
        with open(fn, 'wb') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def __repr__(self):
        return str(self.data)

def _add_item_to_dict(tracking_files, metadata, k, item):
    for fn in tracking_files:
        metadata[fn][k] = item

def _add_items_to_dict(tracking_files, metadata, k, items):
    for fn, item in zip(tracking_files, items):
        metadata[fn][k] = item

def clone_metadata(tracking_files : list, **kwargs) -> dict:
    """
    Prepare a metadata dictionary for defining a VideosetDataFrame. 

    Only required argument is list of DLC tracking file names. 

    Any other keyword argument must be either a non-iterable object (e.g. a scalar parameter, like FPS)
    that will be copied and tagged to each of the DLC tracking files, or an iterable object of the same
    length of the list of DLC tracking files. Each element in the iterable will be tagged with the corresponding
    DLC file.

    Args:
        tracking_files: List of DLC tracking .csvs
        **kwargs: described as above

    Returns:
        Dictionary whose keys are DLC tracking file names, and contains a dictionary with key,values containing the metadata provided
    """

    metadata = {}
    for fn in tracking_files:
        metadata[fn] = {}
    n_files = len(tracking_files)

    for k,v in kwargs.items():
        if hasattr(v, '__len__'):
            if len(v) != n_files:
                _add_item_to_dict(tracking_files, metadata, k, v)
            else:
                _add_items_to_dict(tracking_files, metadata, k, v)
        else:
            _add_item_to_dict(tracking_files, metadata, k, v)

    return metadata

class VideosetDataFrame(MLDataFrame):
    def __init__(self, metadata : dict, 
                       label_key : dict = None, 
                       part_renamer : dict = None,
                       animal_renamer : dict = None):
        """Houses DLC tracking data and behavior annotations in pandas DataFrame for ML, along with relevant metadata, features and behavior annotation labels.

        Args:
            metadata: Dictionary whose keys are DLC tracking csvs, and value is a dictionary of associated metadata
                for that video. Most easiest to create with 'clone_metadata'. 
                Required keys are: ['fps']
            label_key: Default None. Dictionary whose keys are positive integers and values are behavior labels. If none, then this is inferred from the behavior annotation files provided.  
            part_renamer: Default None. Dictionary that can rename body parts from tracking files if needed (for feature creation, e.g.)
            animal_renamer: Default None. Dictionary that can rename animals from tracking files if needed
        """

        self.req_cols = ['fps']

        self.data = pd.DataFrame()
        self.label_key = label_key

        is_valid, should_rescale = self._validate_metadata(metadata)
        if is_valid:   
            self.metadata = metadata
        else:
            raise ValueError("Metadata not properly formatted. See docstring.")
    
        if len(metadata) > 0:
            self._load_tracks(part_renamer, animal_renamer, rescale = should_rescale) 
            self._load_labels(set_as_label = True)
        else:
            self.raw_track_columns = None

        if self.label_key:
            self.reverse_label_key = {v:k for k,v in self.label_key.items()}
        else: 
            self.reverse_label_key = None

        if should_rescale:
            self._convert_units()
        elif 'scale_factor' in self.data.columns: 
            self.data = self.data.drop(columns = 'scale_factor')

        self.feature_cols = None

    def _convert_units(self):
        # if 'frame_width', 'resolution' and 'frame_width_units' are provided, then we convert DLC tracks to these units.
        if len(self.metadata) == 0:
            return 
        for col in self.data.columns:
            is_dlc_feature = False
            for ani in self.animals:
                if col.startswith(ani):
                    is_dlc_feature = True
                    break
            if is_dlc_feature:
                self.data[col] = self.data[col]*self.data['scale_factor']
        self.data = self.data.drop(columns = 'scale_factor')

    def _setup_default_cv_folds(self):
        default_fold_cols = [f'fold{idx}' for idx in range(self.n_videos)]
        self.fold_cols = default_fold_cols 
        raise NotImplementedError

    def _validate_metadata(self, metadata):

        has_all_dim_cols_count = 0

        should_rescale = None
        valid = True

        import numbers
        import warnings

        for fn in metadata:

            n_dim_cols = sum([x in metadata[fn].keys() for x in ['frame_width', 'resolution', 'frame_width_units']])
            if (n_dim_cols > 0) and (n_dim_cols < 3):
                warnings.warn("Must provide all of 'frame_width', 'resolution' and 'frame_width_units' to rescale")
                should_rescale = False
            if (n_dim_cols == 3):
                has_all_dim_cols_count += 1
                if should_rescale is not False:
                    should_rescale = True

                if not isinstance(metadata[fn]['frame_width'], numbers.Number):
                    warnings.warn("'frame_width' must be a number to rescale.")
                    should_rescale = False
                if type(metadata[fn]['frame_width_units']) is not str:
                    warnings.warn("'frame_width_units' must be a string to rescale.")
                    should_rescale = False
                elif metadata[fn]['frame_width_units'].lower() not in UNIT_DICT:
                    warnings.warn(f"Units must be one of the following to rescale: {','.join(UNIT_DICT.keys())}")
                    should_rescale = False

                if hasattr(metadata[fn]['resolution'], '__len__'):
                    if len(metadata[fn]['resolution']) != 2:
                        warnings.warn("'resolution' must be a list-like object of length 2 to rescale.")
                        should_rescale = False

            checks = [col in metadata[fn].keys() for col in self.req_cols]
            if sum(checks) < len(self.req_cols):
                valid = False

        if (has_all_dim_cols_count > 0) and (has_all_dim_cols_count < len(metadata)):
            warnings.warn("Must provide all of 'frame_width', 'resolution' and 'frame_width_units' to rescale")
            should_rescale = False

        if should_rescale is None: should_rescale = False
        if should_rescale:
            print("Rescaling to 'mm'")

        return valid, should_rescale

    @property
    def videos(self):
        return list(self.metadata.keys())

    @property
    def n_videos(self):
        return len(self.metadata)

    @property
    def group(self):
        return self.data.filename.to_numpy()

    def activate_features_by_name(self, name : str) -> list:
        """Add already present columns in data frame to the feature set. 
        
        Args:
            name: string for pattern matching -- any feature that starts with this string will be added

        Returns:
            List of matched columns (may include columns that were already activated).
        """

        matched_cols = [l for l in self.data.columns if re.match(f"^{name}", l)]
        if self.feature_cols is not None:
            self.feature_cols = uniquifier(list(self.feature_cols) + list(matched_cols))
        else:
            self.feature_cols = matched_cols
        return matched_cols

    def get_columns_regex(self, pattern : str) -> list:
        """Return a list of column names that match the provided regex pattern.
        
        Args:
            pattern: a regex pattern to match column names to
            
        Returns:
            list of column names
        """
        try:
            compiled = re.compile(pattern)
        except re.error:
            raise ValueError("Couldn't parse re pattern.")
        matched_cols = [l for l in self.data.columns if compiled.search(l) is not None]
        return matched_cols

    def remove_features_by_name(self, name : str) -> list:
        """Remove columns from the feature set. 
        
        Args:
            name: string for pattern matching -- any feature that starts with this string will be removed

        Returns:
            List of removed columns.
        """
        matched_cols = [l for l in self.feature_cols if re.match(f"^{name}", l)]
        removed_cols = self.remove_feature_cols(matched_cols)
        return removed_cols

    #Set features by individual or by group names
    def add_features(self, feature_maker : Features, 
                           featureset_name : str, 
                           add_to_features = False, 
                           **kwargs) -> list:
        """Compute features to dataframe using Feature object. 'featureset_name' will be prepended to new columns, followed by a double underscore. 

        Args:
            featuremaker: A Feature object that houses the feature-making function to be executed and a list of required columns that must in the dataframe for this to work
            featureset_name: Name to prepend to the added features 
            add_to_features: Whether to add to list of active features (i.e. will be returned by the .features property)
        Returns:
            List of new columns that are computed
        """
        new_features = feature_maker.make(self, **kwargs)

        #Prepend these column names w featureset-name__feature-name
        new_feat_cols = list(new_features.columns)
        new_feat_cols = [str(featureset_name) + '__' + str(i) for i in new_feat_cols]
        new_features.columns = new_feat_cols

        #Don't add duplicated columns:
        notdupcols = [col for col in new_feat_cols if col not in self.data.columns]
        new_features = new_features[notdupcols]

        self.data = pd.concat([self.data.reset_index(drop = True), 
                               new_features.reset_index(drop = True)], axis = 1)
        if add_to_features:
            if self.feature_cols is not None:
                self.feature_cols = list(self.feature_cols) + list(new_features.columns)
            else:
                self.feature_cols = new_features.columns
        return list(new_features.columns)

    def remove_feature_cols(self, col_names : list) -> list:
        """Remove provided columns from set of feature columns.
        
        Args:
            col_names: list of column names
            
        Returns:
            The columns that were removed from those designated as features."""
        new_col_names = [i for i in self.feature_cols if i not in col_names]
        removed = [i for i in self.feature_cols if i in col_names]
        self.feature_cols = new_col_names
        return removed

    def _load_tracks(self, part_renamer, animal_renamer, rescale = False):
        #For the moment only supports DLC
        return self._load_dlc_tracks(part_renamer, animal_renamer, rescale = rescale)

    def _load_dlc_tracks(self, part_renamer, animal_renamer, rescale = False):
        """Add DLC tracks to DataFrame"""
        df = pd.DataFrame()
        dfs = []
        col_names_old = None
        #Go through each video file and load DLC tracks
        for fn in self.metadata.keys():
            df_fn, body_parts, animals, col_names, scorer = read_DLC_tracks(fn, 
                                                                            part_renamer, 
                                                                            animal_renamer)
            n_rows = len(df_fn)
            df_fn['time'] = df_fn['frame']/self.metadata[fn]['fps']
    
            if rescale and ('frame_width_units' in self.metadata[fn]) and \
                    ('frame_width' in self.metadata[fn]) and \
                    ('resolution' in self.metadata[fn]) and \
                    self.metadata[fn]['frame_width_units'].lower() in UNIT_DICT:
                unit_scale_factor = UNIT_DICT[self.metadata[fn]['frame_width_units'].lower()]
                df_fn['scale_factor'] = self.metadata[fn]['frame_width']/self.metadata[fn]['resolution'][1]*unit_scale_factor
                self.metadata[fn]['units'] = 'mm'
            else:
                df_fn['scale_factor'] = 1

            dfs.append(df_fn)
            self.metadata[fn]['duration'] = (n_rows)/self.metadata[fn]['fps']
            self.metadata[fn]['scorer'] = scorer
            if col_names_old is not None:
                if col_names != col_names_old:
                    raise RuntimeError("DLC files have different columns. Must all be from same project")
            col_names_old = col_names

        df = pd.concat(dfs, axis = 0)
        df = df.reset_index(drop = True)
        self.body_parts = body_parts
        self.animals = animals
        self.animal_setup = {'mouse_ids': animals, 'bodypart_ids': body_parts, 'colnames': col_names}
        self.raw_track_columns = col_names
        self.data = df
        for animal in self.animals:
            if animal in ['filename', 'frame', 'time', 'label']:
                print(f"One of the animal names is protected ({animal}), unexpected behavior may occur.")

    def _load_labels(self, col_name = 'label', set_as_label = False):
        #For the moment only BORIS support
        return self._load_labels_boris(col_name, set_as_label)

    def _load_labels_boris(self, col_name = 'label', set_as_label = False):
        """Add behavior label data to DataFrame"""

        if self.label_key is None:
            label_files = []
            for fn in self.metadata.keys():
                if 'label_files' in self.metadata[fn]:
                    label_files.append(self.metadata[fn]['label_files'])
            self.label_key = create_behavior_labels(label_files)

        for vid in self.metadata:
            if 'label_files' in self.metadata[vid]:

                fn_in = self.metadata[vid]['label_files']
                fps = self.metadata[vid]['fps']
                duration = self.metadata[vid]['duration']

                ground_truth, _ = read_boris_annotation(fn_in, fps, duration, self.label_key)

                self.data.loc[self.data['filename'] == vid, col_name] = ground_truth

        if set_as_label:
            self.label_cols = col_name

    def save(self, fn_out : str) -> None:
        """Save VideosetDataFrame object with pickle.
        
        Args:
            fn_out: location to write pickle file to
            
        Returns:
            None. File is saved to path.
        """
        with open(fn_out,'wb') as file:
            file.write(pickle.dumps(self.__dict__, protocol = 4))

    def to_dlc_csv(self, base_dir : str, save_h5_too = False) -> None:
        """Save VideosetDataFrame tracking files to DLC csv format.

        Only save tracking data, not other computed features.
        
        Args:
            base_dir: base_dir to write DLC csv files to
            save_h5_too: if True, also save the data as an h5 file
            
        Returns:
            None. Files are saved to path.
        """
        from itertools import product
        fns = list(self.metadata.keys())
        for fn in fns:
            df = self.data.loc[self.data['filename'] == fn].copy().reset_index(drop = True)
            #Make a multi column object and rearrange table to be in DLC format
            scorer = self.metadata[fn]['scorer']
            selected_cols = []
            for animal, bp, component in product(self.animals, self.body_parts, ['x', 'y', 'likelihood']):
                if component == 'likelihood':
                    old_col_name = f"{component}_{animal}_{bp}"
                else:
                    old_col_name = f"{animal}_{component}_{bp}"
                new_col_name = (scorer, animal, bp, component)
                df = df.rename(columns = {old_col_name: new_col_name})
                selected_cols.append(new_col_name)
            df = df[selected_cols]
            df.columns = pd.MultiIndex.from_tuples(df.columns, names=['scorer', 'individuals', 'bodyparts', 'coords'])
            fn_out = os.path.join(base_dir, os.path.basename(fn))
            df.to_csv(fn_out)
            if save_h5_too:
                df.to_hdf(fn_out.replace('.csv', '.h5'), "df_with_missing", format = 'table', mode="w")

    def load(self, fn_in : str) -> None:
        """Load VideosetDataFrame object from pickle file.
        
        Args:
            fn_in: path to load pickle file from. 
            
        Returns:
            None. Data in this object is populated with contents of file."""
        with open(fn_in, 'rb') as file:
            dataPickle = file.read()
        self.__dict__ = pickle.loads(dataPickle)

    def make_movie(self, label_columns, path_out : str, video_filenames = None) -> None:
        """Given columns indicating behavior predictions or whatever else, make a video
        with these predictions overlaid. 

        VideosetDataFrame metadata must have the keys 'video_file', so that the video associated with each set of DLC tracks is known.

        Args:
            label_columns: list or dict of columns whose values to overlay on top of video. If dict, keys are the columns and values are the print-friendly version.
            path_out: the directory to output the videos too
            video_filenames: list or string. The set of videos to use. If not provided, then use all videos as given in the metadata.

        Returns:
            None. Videos are saved to 'path_out'
        """

        if not checkFFMPEG():
            print("Cannot find ffmpeg in path. Please install.")
            return

        #Text parameters (units in pixels)
        y_offset = global_config['make_movie__y_offset']
        y_inc = global_config['make_movie__y_inc']
        text_color = global_config['make_movie__text_color']
        font_size = global_config['make_movie__font_size']

        if type(video_filenames) is str:
            video_filenames = [video_filenames]
        if not video_filenames:
            video_filenames = self.videos
        if type(label_columns) is list:
            label_columns = {k:k for k in label_columns}

        for video in video_filenames:
            rate = 1/self.metadata[video]['fps']
            vid_in = self.metadata[video]['video_files']
            file_out = os.path.splitext(os.path.basename(vid_in))[0] + '_'.join(label_columns.keys()) + '.mp4'
            vid_out = os.path.join(path_out, file_out)
            label_strings = []
            for idx, (col,print_label) in enumerate(label_columns.items()):
                this_y_offset = y_offset + y_inc*idx
                behavior_pairs = _make_dense_values_into_pairs(self.data.loc[self.data.filename == video, col], rate)
                this_label_string = ','.join([f"drawtext=text=\'{print_label}\':x=90:y={this_y_offset}:fontsize={font_size}:fontcolor={text_color}:enable=\'between(t,{pair[0]},{pair[1]})\'" for pair in behavior_pairs])
                label_strings.append(this_label_string)
            label_string = ','.join(label_strings)

            #ffmpeg is the fastest way to add this information to a video
            #Prepare the ffmpeg command
            cmd = f'ffmpeg -y -i {vid_in} -vf "{label_string}" {vid_out}'
            os.system(cmd)
            
def load_videodataset(fn_in : str) -> VideosetDataFrame:
    """Load VideosetDataFrame from file.
    
    Args:
        fn_in: path to file to load
        
    Returns:
        VideosetDataFrame object from pickle file
    """
    with open(fn_in, 'rb') as file:
        dataPickle = file.read()
    new_obj = VideosetDataFrame({})
    new_obj.__dict__ = pickle.loads(dataPickle)
    return new_obj

def get_sample_openfield_data():
    """Load a sample dataset of 1 mouse in openfield setup. The video is the sample that comes with DLC.
    
    Returns:
        (VideosetDataFrame) Data frame with the corresponding tracking and behavior annotation files
    """

    cur_dir = os.path.dirname(os.path.abspath(__file__))
    tracking_files = glob(os.path.join(cur_dir, 'data', 'dlc', 'openfield', '*example.csv'))
    video_files = glob(os.path.join(cur_dir, 'data', 'videos', '*.mp4'))
    fps = 30                         # (int) frames per second
    resolution = (480, 640)        # (tuple) HxW in pixels
    #Metadata is a dictionary that attaches each of the above parameters to the video/behavior annotations
    metadata = clone_metadata(tracking_files, 
                            video_files = video_files,
                            fps = fps, 
                            resolution = resolution)

    vdf = VideosetDataFrame(metadata)

    return vdf

def _make_dense_values_into_pairs(predictions, rate):
    #Put into start/stop pairs
    pairs = []
    in_pair = False
    start = 0
    for idx, behavior in enumerate(predictions):
        if behavior == 0:
            if in_pair:
                pairs.append([start, idx*rate])
            in_pair = False
        if behavior == 1:
            if not in_pair:
                start = idx*rate
            in_pair = True

    if in_pair:
        pairs.append([start, idx*rate])
    return pairs
    


""" Functions to take pose tracks and compute a set of features from them.

To make your own feature creator:

Create a function, e.g. `create_custom_features`, and provide the Features class a list of columns that are needed by this function to compute the features.

The function `create_custom_features` has the form:

`create_custom_features(<df>, <raw_col_names>, <animal_setup>, **kwargs)`

Where:

`df` is the dataframe to compute the features on
`raw_col_names` is a list of the names of the columns in the dataframe that contain the raw data used for the feature creation. These are required for the model.
`animal_setup` is a dictionary with keys `bodypart_ids`, `mouse_ids`, `colnames`.
   `bodypart_ids` is a list of the bodypart ids that are used in the dataframe
   `mouse_ids` is a list of the mouse ids that are used in the dataframe
   `colnames` is the list product(animals, XY_IDS, body_parts) 
`**kwargs` are extra arguments passed onto the feature creation function.

The function returns:

A dataframe, that only contains the new features. These will be added to the VideosetDataFrame as columns.

Once you have such a function defined, you can create a "feature making object" with

`custom_feature_maker = Features(create_custom_features, req_columns)`

This could be used on datasets as:

```
dataset.add_features(custom_feature_maker, featureset_name = 'CUSTOM', add_to_features = True)
```
"""

from typing import Callable
import warnings

from behaveml.dl.dl_features import compute_dl_probability_features
from behaveml.mars_features import compute_mars_features, compute_mars_reduced_features, compute_social_features
from behaveml.generic_features import compute_centerofmass_interanimal_distances, \
                                        compute_centerofmass_interanimal_speed, \
                                        compute_centerofmass, \
                                        compute_centerofmass_velocity, \
                                        compute_speed_features, \
                                        compute_distance_features

default_tracking_columns = ['resident_x_nose', 'resident_x_leftear', 'resident_x_rightear', 'resident_x_neck',
                            'resident_x_lefthip', 'resident_x_righthip', 'resident_x_tail', 'resident_y_nose',
                            'resident_y_leftear', 'resident_y_rightear', 'resident_y_neck',
                            'resident_y_lefthip', 'resident_y_righthip', 'resident_y_tail',
                            'intruder_x_nose', 'intruder_x_leftear', 'intruder_x_rightear',
                            'intruder_x_neck', 'intruder_x_lefthip', 'intruder_x_righthip',
                            'intruder_x_tail', 'intruder_y_nose', 'intruder_y_leftear',
                            'intruder_y_rightear', 'intruder_y_neck', 'intruder_y_lefthip',
                            'intruder_y_righthip', 'intruder_y_tail']

class Features(object):
    def __init__(self, feature_maker : Callable, required_columns : list, **kwargs):
        """Feature creation object. This houses the feature creation function and the columns that are required to compute the features. Performs some checks on data to make sure has these columns.

        See docstring for the `features` model for more information.

        Args:
            feature_maker: The function that will be used to compute the features.
            required_columns: The columns that are required to compute the features.
        """
        self.required_columns = required_columns
        self.feature_maker = feature_maker
        self.kwargs = kwargs

    def make(self, vdf, **kwargs):
        """Make the features. This is called internally by the dataset object when running `add_features`.

        Args:
            vdf: The VideosetDataFrame to compute the features on.
            **kwargs: Extra arguments passed onto the feature creation function.
        """
        #Validate columns:
        checks = [col in self.required_columns for col in vdf.data.columns]
        if sum(checks) < len(self.required_columns):
            raise RuntimeError("VideosetDataFrame doesn't have necessary columns to compute this set of features.")
        if vdf.data[self.required_columns].isnull().values.any():
            warnings.warn("Missing values in required data columns. May result in unexpected behavior. Consider interpolating or imputing missing data first.")
        new_features = self.feature_maker(vdf.data, self.required_columns, vdf.animal_setup, **self.kwargs, **kwargs)
        return new_features

## MARS features
mars_feature_maker = Features(compute_mars_features, default_tracking_columns)
marsreduced_feature_maker = Features(compute_mars_reduced_features, default_tracking_columns)
cnn_probability_feature_maker = Features(compute_dl_probability_features, default_tracking_columns)
social_feature_maker = Features(compute_social_features, default_tracking_columns)

## Generic features -- don't need any specific column names. Will be based on the animal setup.
com_interanimal_feature_maker = Features(compute_centerofmass_interanimal_distances, [])
com_interanimal_speed_feature_maker = Features(compute_centerofmass_interanimal_speed, [])
com_feature_maker = Features(compute_centerofmass, [])
com_velocity_feature_maker = Features(compute_centerofmass_velocity, [])
speed_feature_maker = Features(compute_speed_features, [])
distance_feature_maker = Features(compute_distance_features, [])


import pandas as pd 
import numpy as np
from behaveml.video import VideosetDataFrame

def interpolate_lowconf_points(vdf : VideosetDataFrame,
                               conf_threshold : float = 0.9,
                               in_place : bool = True,
                               rolling_window : bool = True,
                               window_size : int = 3) -> pd.DataFrame:
    """Interpolate raw tracking points if their probabilities are available.

    Args:
        vdf: VideosetDataFrame containing the tracks to interpolate
        conf_threshold: default 0.9. Confidence below which to count as uncertain, and to interpolate its value instead
        in_place: default True. Whether to replace data in place
        rolling_window: default True. Whether to use a rolling window to interpolate
        window_size: default 3. The size of the rolling window to use

    Returns:
        Pandas dataframe with the filtered raw columns. Returns None if opted for in_place modification
    """

    df_filtered = []

    for fn_in in vdf.videos:
        print("processing", fn_in)

        if not in_place:
            df_filter_low_conf = vdf.data.loc[vdf.data.filename == fn_in].copy()
        else:
            df_filter_low_conf = vdf.data
        
        for m in vdf.animals:
            for bp in vdf.body_parts:
                low_conf = df_filter_low_conf.loc[vdf.data.filename == fn_in, '_'.join(['likelihood', m, bp])] < conf_threshold
                df_filter_low_conf.loc[(vdf.data.filename == fn_in) & low_conf,'_'.join([m, 'x', bp])] = np.nan
                df_filter_low_conf.loc[(vdf.data.filename == fn_in) & low_conf,'_'.join([m, 'y', bp])] = np.nan
                
        df_filter_low_conf.loc[(vdf.data.filename == fn_in)] = \
            df_filter_low_conf.loc[(vdf.data.filename == fn_in)].\
            interpolate(axis = 0, method = 'linear', limit_direction = 'both')            

        if rolling_window:
            df_filter_low_conf.loc[(vdf.data.filename == fn_in), vdf.raw_track_columns] = \
                df_filter_low_conf.loc[(vdf.data.filename == fn_in), vdf.raw_track_columns].rolling(window = window_size, min_periods = 1).mean()
            
        df_filter_low_conf = df_filter_low_conf[vdf.raw_track_columns + ['filename', 'frame']]
        df_filtered.append(df_filter_low_conf)

    df_filtered = pd.concat(df_filtered, axis = 0)

    if not in_place:
        return df_filtered
    else:
        return None

#TODO
# Get this working...
# if filter_out_toofast:
#     for m in vdf.animals:
#         for bp in vdf.body_parts:
#             too_quick = (df_filter_low_conf.loc[:,'_'.join([m, 'x', bp])].diff(jump_dur).abs() > speed_threshold) | \
#                         (df_filter_low_conf.loc[:,'_'.join([m, 'y', bp])].diff(jump_dur).abs() > speed_threshold) | \
#                         (df_filter_low_conf.loc[:,'_'.join([m, 'x', bp])].diff(-jump_dur).abs() > speed_threshold) | \
#                         (df_filter_low_conf.loc[:,'_'.join([m, 'y', bp])].diff(-jump_dur).abs() > speed_threshold)

#             df_filter_low_conf.loc[too_quick,'_'.join([m, 'x', bp])] = np.nan
#             df_filter_low_conf.loc[too_quick,'_'.join([m, 'y', bp])] = np.nan

#     #And now filter again
#     df_filter_low_conf = df_filter_low_conf.interpolate(axis = 0, method = 'polynomial', order = 1)
#     df_filter_low_conf = df_filter_low_conf.rolling(window = 3, min_periods = 1).mean()

#Try using the fancyimpute package....
# #df_filter_low_conf = pd.DataFrame(NuclearNormMinimization().fit_transform(df_filter_low_conf.to_numpy()), 
# #                                 index = df_filter_low_conf.index, 
# #                                 columns = df_filter_low_conf.columns)

# df_filter_low_conf = pd.DataFrame(KNN(k=7).fit_transform(df_filter_low_conf.to_numpy()), 
#                                 index = df_filter_low_conf.index, 
#                                 columns = df_filter_low_conf.columns)

# #X_incomplete_normalized = BiScaler().fit_transform(df_filter_low_conf.to_numpy())
# #df_filter_low_conf = pd.DataFrame(SoftImpute().fit_transform(X_incomplete_normalized),
# #                                 index = df_filter_low_conf.index,
# #                                 columns = df_filter_low_conf.columns)



"""Configuration options for behaveml functions."""

#TODO
#Add support for the user to change these.

global_config = {
    'make_movie__y_offset': 60,
    'make_movie__y_inc': 30,
    'make_movie__text_color': 'green',
    'make_movie__font_size': 36,
    'create_mosaic_video__max_mosaic_vids': 16
}
